#ifndef RAWMICROSLICE_H
#define RAWMICROSLICE_H
#include "DAQDataFormats/FEBSimulator.h"
#include "DAQDataFormats/RawMicroSliceHeader.h"
#include "DAQDataFormats/RawTimingMarker.h"
#include "DAQDataFormats/RawNanoSlice.h"
#include "DAQDataFormats/Macros.h"
#include "DAQDataFormats/FunctionBind.h"

#include <vector>		/* Use the Vector STL */



  //! Class to hold the data from the FEBs in correct formats
  /*!
    The RawMicroSlice class is designed to mimic the format of the data
   * produced in the DCM by the DCM firmware
   *
   *
   * CAUTION!!!!!!!!!!! Right now the MicroSlice Header size is considered to be equal to 1
   * always! If you plan on changing it, could be a hassle.
   * RawMicroSlice::setBufferSource uses 1 explisitly
   *
  */



namespace daqdataformats {

/*
// FLAG Setting for Initialization and Resets
#define MICROSLICE_RESET_MEM 0x1
#define MICROSLICE_RESET_FEB 0x2
#define MICROSLICE_DELETE_FEB 0x4
#define MICROSLICE_INIT_FEB  0x8
*/
    namespace rawmicro{
        static const uint32_t DEFAULT_MICROSLICE_SIZE       = 16384;  /* Maximum Microslice Buffer Size 64k/4 */
        static const uint32_t MAXIMUM_FRONTEND_BOARDS       = 64;     /* Maximum number of front end boards we support */
        static const uint32_t PREDETERMINED_MICROSLICE_SIZE = RAWMICROSLICEHEADER_SIZE + RAWTIMINGMARKER_SIZE;
    }

   //! Class to hold the MicroSlice data
   /*!
   Currently, MicroSlice does not have versioning
   */

    class RawMicroSlice : public RawDAQData {
      

        public:
    
            // Constructors
            /// If resize_buffer=true, then we resize the whole internal Buffer with zeros on the construction
            RawMicroSlice(const uint32_t  buffersize               = rawmicro::DEFAULT_MICROSLICE_SIZE
                         ,const bool      resize_buffer            = false
                         ,const version_t microsliceheader_version = rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION
                         );
            RawMicroSlice(const RawMicroSlice& copy_in);//! Deep copy constructor
            RawMicroSlice(const void*     microslice_buffer
                         ,const uint32_t  buffersize               = rawmicro::DEFAULT_MICROSLICE_SIZE
                         ,const version_t microsliceheader_version = rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION
                         ); //! Shallow copy constructor
            /// Destructor
            ~RawMicroSlice();

            version_t getVersion() const {return _microSliceHeaderVersion;}
            bool      setVersion(const version_t);

            /// Add a NanoSlice to the MicroSlice changing the MicroSlice Header appropriately
            /// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
            /// and adds an appropriate flag to the MicroSlice Header.
            void      addNanoSlice(const void* nanoslice_buffer, const bool check_mc_flag=true);
            /// Add a Some nwords of data to the internal buffer changing the MicroSlice Header appropriately
            /// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
            /// and adds an appropriate flag to the MicroSlice Header.
            void      addData(const void* buffer, const uint32_t nwords, const bool check_mc_flag=true);
    
            /// Generate a MicroSlice
            /// Need to provide the version of NanoSlices to generate
            void      generate(const version_t version_nanoslice=nanosliceheader::RAWNANOSLICE_DEFAULT_VERSION);

            // Getters
            daqdataformats::RawMicroSliceHeader* getHeader() const;               //! Get the MicroSlice Header
            daqdataformats::RawTimingMarker*     getTimingMarker() const;         //! Get the MicroSlice Timing Marker
            daqdataformats::RawNanoSlice*        getFloatingNanoSlice() const;    //! Get Floating NanoSlice
            bool      getDataPresent()    const; //! Get the Data Present bit of the Header
            bool      getMonteCarloFlag() const; //! Get the Data Present bit of the Header
            uint32_t  getByteCount()      const; //! Get the byte count value of the header
            uint32_t  getLowWord()        const; //! Get method for lower half of timing word
            uint32_t  getHighWord()       const; //! Get method for upper half of timing word
            uint64_t  getTime()           const; //! Get the time-stamp as a single 64-bit value
            float     getHitProbability() const; //! Get current Hit Probability (for a microslice generation)
            uint32_t  getNumNanoSlices()  const; //! Get number of NanoSlices in the MicroSlice

            //Setters
            bool      setMonteCarloFlag    (const bool);            //! Set the SIM flag
            bool      setFloatingNanoSlice (const uint32_t inano) const; //! Set Floating NanoSlice to point to the "inano"-th NanoSlice
                                                              /// We are allowed to change the floating nanoslice, thus it's const
            bool      setLowWord           (const uint32_t);	 //! Set method for lower half of timing word
            bool      setHighWord          (const uint32_t);	 //! Set method for upper half of timing word
            bool      setTime              (const uint64_t); /**< Set the time-stamp as a single 64-bit value */
            /// Set Hit probability of a pixel. 
            /// Used when generating MicroSlices - method generate.
            bool      setHitProbability    (const float hit_prob);

            /// Getting a nanoslice size. It could be different for different versions
            uint32_t  getNanoSliceSize() const;

            using          RawDAQData::readData;
            /// Read Data From File Descriptor
            readDataStatus readData(const int file_descriptor);

            void      init            ();
            uint32_t  sizeofdata      () const;
            bool      setBufferSource (const void* pointer);

        protected:

            // These have to have a different name from readData and print.
            // Otherwise, we can't bind them to their respactive DELEGATEs.
            void*     readDataFromBuffer(const void*);
            void      printStandard     (std::ostream& os=std::cout) const;

            /// Sometimes we don't know the NanoSlice size yet,
            /// But we need to know if there are NanoSlices at all
            bool      areThereNanoSlices() const;

            /// Read the NanoSlice version from the buffer
            version_t figureOutTheVersion(const void*) const{return 0;}

            uint32_t  getPredeterminedSize() const { return rawmicro::PREDETERMINED_MICROSLICE_SIZE;}
            bool      setFunctionPointers(){return true;};

        private:

            daqdataformats::RawMicroSliceHeader*  _Header;           ///< MicroSlice Header
            daqdataformats::RawTimingMarker*      _TimingMarker;     ///< MicroSlice Timing Marker
            mutable daqdataformats::RawNanoSlice* _FloatingNanoSlice;///< Floating NanoSlice.
                                                                     ///< Allowed to be changed anywhere, even in const routines
                                                                     ///< One can set the pointer of the
                                                                     ///< nanoslice to any one within the MicroSlice
            float                                 _HitProbability;   ///< Pixel Hit probability. Used when generating a MicroSlice
            mutable version_t                     _nanosliceVersion;
            mutable version_t                     _microSliceHeaderVersion;
    };// end of class RawMicroSlice
    
} // END NAMESPCE daqdataformats
#endif /* RAWMICROSLICEHEADER_H */
