#ifndef TRIGGERDEFINES_H
#define TRIGGERDEFINES_H

#include <sys/types.h>

//! This header contains all of the enumerated data that specify the trigger types
// See https://cdcvs.fnal.gov/redmine/documents/253 for more info

namespace daqdataformats{

  //! Trigger Source Enumeration
  enum trigSource {
    // These are the source type that is issuing the trigger

    /*************************************************************************
     * Standard Systems (these are predefined and guaranteed not to change   *
     *************************************************************************/
    // Beam Time Stamp Systems (i.e. tdu master)
    TRIGSRC_BEAMSTAMP_SYSTEM       =1,
    // Pulser System (Software)
    TRIGSRC_PULSER_SYSTEM          =2,
    // Pulser System (Hardware)
    TRIGSRC_HARDWARE_PULSER_SYSTEM =3,
    // Software Trigger from Buffer Nodes
    TRIGSRC_BUFFNODE_SYSTEM        =4,
    // Trigger Originating from Global Trigger System (e.g. software pulser)
    TRIGSRC_GT_SYSTEM              =5,
    // Manual Triggering of the event
    TRIGSRC_MANUAL                 =6,
    // Simulated Event Play back trigger
    TRIGSRC_SIMULATED              =7,
    // External SNEWS trigger
    TRIGSRC_SNEWS     		   =8,
    /*************************************************************************
     * User Defined Sources (these should be named when they are first used) *
     *************************************************************************/
    TRIGSRC_USERDEF_1 =8,
    TRIGSRC_USERDEF_2 =9,
    TRIGSRC_USERDEF_3 =10,
    TRIGSRC_USERDEF_4 =11,
    TRIGSRC_USERDEF_5 =12,
    TRIGSRC_USERDEF_6 =13,
    /*************************************************************************
     * Catch all sources (if you don't have something defined yet)           *
     *************************************************************************/
    // Unknown Source 
    TRIGSRC_UNKNOWN   = 14,
    // Dummy Source for testing
    TRIGSRC_DUMMY     = 0xF,
    TRIGSRC_MAX       = 0xF
  };

  enum trigSourceID{
    // Beam Time Stamp Source (i.e. TDU Masters) by location
    TRIGSRC_ID_BEAMSTAMP_NUMI_1    = 1,              // Location: NuMI Cavern
    TRIGSRC_ID_BEAMSTAMP_NUMI_2,                     // same, but from kNuMItclk instead of just kNuMI
    TRIGSRC_ID_BEAMSTAMP_NDOS_1    = 10,             // Location: Near Detector on the Surface Building
    TRIGSRC_ID_BEAMSTAMP_NDOS_2,
    TRIGSRC_ID_BEAMSTAMP_NDOS_3,
    TRIGSRC_ID_BEAMSTAMP_NDOS_4,
    TRIGSRC_ID_BEAMSTAMP_MINOS_1   = 20,             // Location: Minos Service Building
    TRIGSRC_ID_BEAMSTAMP_MINOS_2,
    TRIGSRC_ID_BEAMSTAMP_FEYNMAN_1 = 30,             // Location: FCC3
    TRIGSRC_ID_BEAMSTAMP_FEYNMAN_2,
    TRIGSRC_ID_BEAMSTAMP_BOOSTER_1 = 40,             // Location: MI-12
    TRIGSRC_ID_BEAMSTAMP_BOOSTER_2,                  // same, but from kBNBtclk instead of just kBNB
    TRIGSRC_ID_BEAMSTAMP_PPS       = 50,             // from the accelerator PPS signal
    TRIGSRC_ID_BEAMSTAMP_UNKNOWN   = 98,
    TRIGSRC_ID_BEAMSTAMP_MAX       = 99,             // Maximum ID for a TDU source 

    // Pulser Source Block
    TRIGSRC_ID_PULSER_P1   =200,  // cal_per
    TRIGSRC_ID_PULSER_P2,         // cal_rnd
    TRIGSRC_ID_PULSER_P3,         // stream_per
    TRIGSRC_ID_PULSER_P4,         // stream_rnd
    TRIGSRC_ID_PULSER_P5,
    TRIGSRC_ID_PULSER_P6,
    TRIGSRC_ID_PULSER_P7,
    TRIGSRC_ID_PULSER_P8,
    TRIGSRC_ID_PULSER_P9,
    TRIGSRC_ID_PULSER_P10,
    TRIGSRC_ID_PULSER_P11,
    TRIGSRC_ID_PULSER_P12,
    TRIGSRC_ID_PULSER_UNKNOWN  = 298,
    TRIGSRC_ID_PULSER_MAX      = 299,

    // Global Trigger Source Block
    TRIGSRC_ID_GT_FAR      = 300,
    TRIGSRC_ID_GT_NEAR,
    TRIGSRC_ID_GT_NDOS,
    TRIGSRC_ID_GT_TEST,
    TRIGSRC_ID_GT_UNKNOWN  = 398,
    TRIGSRC_ID_GT_MAX      = 399,
    TRIGSRC_ID_GT_DUMMY    = 399,

    // SNEWS Trigger Sources
    TRIGSRC_ID_SNEWS_BNL       = 400,
    TRIGSRC_ID_SNEWS_MASTER    = 400,
    TRIGSRC_ID_SNEWS_INFN      = 401, 
    TRIGSRC_ID_SNEWS_SECONDARY = 401, 

    // Buffer Node Sources
    TRIGSRC_ID_BUFFNODE_1       = 1000,
    TRIGSRC_ID_BUFFNODE_2,
    TRIGSRC_ID_BUFFNODE_3,
    TRIGSRC_ID_BUFFNODE_4,
    TRIGSRC_ID_BUFFNODE_5,
    TRIGSRC_ID_BUFFNODE_6,
    TRIGSRC_ID_BUFFNODE_7,
    TRIGSRC_ID_BUFFNODE_8,
    TRIGSRC_ID_BUFFNODE_9,
    TRIGSRC_ID_BUFFNODE_10,
    TRIGSRC_ID_BUFFNODE_11,
    TRIGSRC_ID_BUFFNODE_12,
    TRIGSRC_ID_BUFFNODE_13,
    TRIGSRC_ID_BUFFNODE_14,
    TRIGSRC_ID_BUFFNODE_15,
    TRIGSRC_ID_BUFFNODE_16,
    TRIGSRC_ID_BUFFNODE_17,
    TRIGSRC_ID_BUFFNODE_18,
    TRIGSRC_ID_BUFFNODE_19,
    TRIGSRC_ID_BUFFNODE_20,
    TRIGSRC_ID_BUFFNODE_UNKNOWN = 1998,
    TRIGSRC_ID_BUFFNODE_DUMMY   = 1999,
    TRIGSRC_ID_BUFFNODE_MAX     = 1999,

    TRIGSRC_ID_UNKNOWN          = 0xFFFE,
    TRIGSRC_ID_DUMMY            = 0xFFFF,
    TRIGSRC_ID_MAX              = 0xFFFF

  };

  enum trigSourceSubID{
    // These are the specific units or some other sub-ID name
    // to fully enumerate a given source of a trigger
    // these need not have unique values outside of their
    // given source class (i.e. TDU-01 could have the same
    // number as DCM-01 since both have different parent
    // classes.  The maximum value of this field is 0xFF.
    TRIGSRC_SUBID_TDU_01 = 1,
    TRIGSRC_SUBID_TDU_02,
    TRIGSRC_SUBID_TDU_03,
    TRIGSRC_SUBID_TDU_04,
    TRIGSRC_SUBID_TDU_05,
    TRIGSRC_SUBID_TDU_06,
    TRIGSRC_SUBID_TDU_07,
    TRIGSRC_SUBID_TDU_08,

    TRIGSRC_SUBID_PULSER_C1=100,
    TRIGSRC_SUBID_PULSER_C2,
    TRIGSRC_SUBID_PULSER_C3,
    TRIGSRC_SUBID_PULSER_C4,
    TRIGSRC_SUBID_PULSER_C5,
    TRIGSRC_SUBID_PULSER_C6,
    TRIGSRC_SUBID_PULSER_C7,
    TRIGSRC_SUBID_PULSER_C8,

    TRIGSRC_SUBID_PARTITION_0=200,
    TRIGSRC_SUBID_PARTITION_1,
    TRIGSRC_SUBID_PARTITION_2,
    TRIGSRC_SUBID_PARTITION_3,
    TRIGSRC_SUBID_PARTITION_4,

    TRIGSRC_SUBID_SNEWS_TRIG=300,
    TRIGSRC_SUBID_SNEWS_BEAT_SLOW,
    TRIGSRC_SUBID_SNEWS_BEAT_FAST,

    TRIGSRC_SUBID_OTHER_DETECTOR,

    TRIGSRC_SUBID_DUMMY = 0xFF
  };

  enum trigClockSource{
    // These are the clock sources which caused the trigger
    // This corresponds to the system which ISSUED the 
    // initial trigger (i.e. for beam spill triggers this
    // is typically a TDU's clock, while for a software
    // pulser this would be the Unix system clock)

    // GPS Based Clocks (Extend from 1-99)
    CLOCKSRC_GPS_FNAL_NUMI_1 = 1,
    CLOCKSRC_GPS_FNAL_NUMI_2,
    CLOCKSRC_GPS_FNAL_NUMI_3,

    CLOCKSRC_GPS_FNAL_NDOS_1 = 10,
    CLOCKSRC_GPS_FNAL_NDOS_2,
    CLOCKSRC_GPS_FNAL_NDOS_3,

    CLOCKSRC_GPS_FNAL_MINOS_1 = 20,
    CLOCKSRC_GPS_FNAL_MINOS_2,
    CLOCKSRC_GPS_FNAL_MINOS_3,

    CLOCKSRC_GPS_FNAL_FEYNMAN_1 = 30,
    CLOCKSRC_GPS_FNAL_FEYNMAN_2,

    CLOCKSRC_GPS_FNAL_BOOSTER_1 = 40,
    CLOCKSRC_GPS_FNAL_BOOSTER_2,

    CLOCKSRC_GPS_ASHRIVER_1 = 50,
    CLOCKSRC_GPS_ASHRIVER_2,

    // Unix System Clocks
    CLOCKSRC_UNIX_PPC_FNAL_NUMI_1 = 1,
    CLOCKSRC_UNIX_PPC_FNAL_NUMI_2,
    CLOCKSRC_UNIX_PPC_FNAL_NUMI_3,

    CLOCKSRC_UNIX_PPC_FNAL_NDOS_1 = 10,
    CLOCKSRC_UNIX_PPC_FNAL_NDOS_2,
    CLOCKSRC_UNIX_PPC_FNAL_NDOS_3,

    CLOCKSRC_UNIX_PPC_FNAL_MINOS_1 = 20,
    CLOCKSRC_UNIX_PPC_FNAL_MINOS_2,
    CLOCKSRC_UNIX_PPC_FNAL_MINOS_3,

    CLOCKSRC_UNIX_PPC_FNAL_FEYNMAN_1 = 30,
    CLOCKSRC_UNIX_PPC_FNAL_FEYNMAN_2,

    CLOCKSRC_UNIX_PPC_FNAL_BOOSTER_1 = 40,
    CLOCKSRC_UNIX_PPC_FNAL_BOOSTER_2,

    CLOCKSRC_UNIX_X86_FNAL_GT_1     = 100,
    CLOCKSRC_UNIX_X86_FNAL_GT_2,
    CLOCKSRC_UNIX_X86_FNAL_GT_3,

    CLOCKSRC_UNIX_X86_ASHRIVER_GT_1     = 200,
    CLOCKSRC_UNIX_X86_ASHRIVER_GT_2,

    CLOCKSRC_UNIX_X86_FNAL_BUFFERFARM_1 = 1000,
    CLOCKSRC_UNIX_X86_FNAL_BUFFERFARM_2,

    CLOCKSRC_UNIX_X86_ASHRIVER_BUFFERFARM_1 = 2000,
    CLOCKSRC_UNIX_X86_ASHRIVER_BUFFERFARM_2,
 
    CLOCKSRC_UNKNOWN,
    CLOCKSRC_DUMMY = 0xFF,
    CLOCKSRC_MAX = 0xFF
  };

#define MAX_TRIGGER_BITS 88

  // Once defined these are forever set in stone
  // DO NOT MODIFY ANY TRIGGER VALUE IN THIS BLOCK
  enum trigID{
    TRIG_ID_NULL            = 0,
    TRIG_ID_BEAM_NUMI       = 1,
    TRIG_ID_BEAM_BOONE      = 2,
    TRIG_ID_SOFT_CAL_PER    = 3,
    TRIG_ID_SOFT_CAL_RND    = 4,
    TRIG_ID_SOFT_MANUAL     = 5,
    TRIG_ID_BEAM_1PPS       = 6,
    TRIG_ID_SOFT_STREAM_PER = 7,
    TRIG_ID_SOFT_STREAM_RND = 8,
    TRIG_ID_SOFT_DAQ_STATUS = 9,

    // Data Driven ones
    TRIG_ID_DATA_SN         = 13,
    TRIG_ID_DATA_ENERGY     = 14,
    TRIG_ID_DATA_CALMU      = 15,
    TRIG_ID_DATA_UPMU       = 16, 
    TRIG_ID_DATA_NUMU       = 17,
    TRIG_ID_DATA_NUE        = 18,
    TRIG_ID_DATA_FASTMONO   = 19,
    TRIG_ID_DATA_SLOWMONO   = 20,

    TRIG_ID_DATA_FEBOVERFLOW= 21,
    TRIG_ID_DATA_ACTIVITY_1 = 22,
    TRIG_ID_DATA_ACTIVITY_2 = 23,

    TRIG_ID_SOFT_MONTECARLO = 24,

    TRIG_ID_SNEWS_TRIGGER   = 25,
    TRIG_ID_SNEWS_BEAT_SLOW = 26,
    TRIG_ID_SNEWS_BEAT_FAST = 27,
    TRIG_ID_DATA_CONTAINED  = 28,
    TRIG_ID_DATA_MOON       = 29,
    TRIG_ID_DATA_SUN        = 30,
    TRIG_ID_DATA_MICHEL     = 31,
    TRIG_ID_DATA_SNXTRIG    = 32,
    TRIG_ID_DATA_NEW1       = 33,
    TRIG_ID_DATA_NEW2       = 34,
    TRIG_ID_DATA_MICHEL_VAR2 = 35,
    TRIG_ID_DATA_MICHEL_VAR3 = 36,
    TRIG_ID_DATA_MICHEL_VAR4 = 37,
    TRIG_ID_DATA_MICHEL_TEST = 38,
    TRIG_ID_DATA_FASTMONONN = 39,
    TRIG_ID_DATA_FASTMONONN_VAR2 = 40,
    TRIG_ID_DATA_FASTMONONN_VAR3 = 41,
    TRIG_ID_DATA_FASTMONONN_TEST = 42,
    TRIG_ID_MAX             = 88
  };

  enum trigBitID{
    // Precomputed Trigger Masks
    TRIG_BIT_NULL            =0x00,
    TRIG_BIT_BEAM_NUMI       =0x01,     // A NuMI beam trigger
    TRIG_BIT_BEAM_BOONE      =0x02,     // A Booster beam trigger
    TRIG_BIT_SOFT_CAL_PER    =0x04,     // A calibration periodic trigger
    TRIG_BIT_SOFT_CAL_RND    =0x08,     // A calibration random trigger
    TRIG_BIT_SOFT_MANUAL     =0x10,     // A manually issued trigger
    TRIG_BIT_BEAM_1PPS       =0x20,     // An accelerator 1 pulse/sec trigger
    TRIG_BIT_SOFT_STREAM_PER =0x40,     // A streamer periodic trigger
    TRIG_BIT_SOFT_STREAM_RND =0x80,     // A streamer random trigger
    TRIG_BIT_SOFT_DAQ_STATUS =0x100,    // A DAQ status block trigger
    TRIG_BIT_SOFT_MONTECARLO =0x800000, // A Montecarlo Trigger

    /***********************************************************************
     * Data Driven Triggers, one for each algorithm.                       *
     * There is also room in the "trigger mod" field in the trigger header *
     * to encode further info, eg, what version of the algorithm           *
     ***********************************************************************/
    TRIG_BIT_DATA_SN         =0x001000,     // a SN trigger
    TRIG_BIT_DATA_ENERGY     =0x002000,     // big-energy events (brehms, air showers)
    TRIG_BIT_DATA_CALMU      =0x004000,     // a calibration muon trigger
    TRIG_BIT_DATA_UPMU       =0x008000,     // upgoing muon trigger
    TRIG_BIT_DATA_NUMU       =0x010000,     // contained nu-induced muons
    TRIG_BIT_DATA_NUE        =0x020000,     // electron showers
    TRIG_BIT_DATA_FASTMONO   =0x040000,     // fast monopoles
    TRIG_BIT_DATA_SLOWMONO   =0x080000,     // slow monopoles
    TRIG_BIT_DATA_FEBOVERFLOW=0x100000,     // FEB Overflow error trigger
    TRIG_BIT_DATA_ACTIVITY_1 =0x200000,     // Activity Trigger Variant 1 (Near Det)
    TRIG_BIT_DATA_ACTIVITY_2 =0x400000,     // Activity Trigger Variant 2 (Near Det)

    // warning: SNEWS Trigger on up uses mid word
    TRIG_BIT_SNEWS_TRIGGER          =0x00000001,   // SNEWS Trigger, mid word
    TRIG_BIT_SNEWS_BEAT_SLOW        =0x00000002,   // SNEWS Heart Beat Slow, mid word
    TRIG_BIT_SNEWS_BEAT_FAST        =0x00000004,   // SNEWS Heart Beat Fast, mid word
    TRIG_BIT_DATA_CONTAINED         =0x00000008,   // a contained vertex trigger, mid word
    TRIG_BIT_DATA_MOON              =0x00000010,   // muon from near moon, mid word
    TRIG_BIT_DATA_SUN               =0x00000020,   // muon from near sun, mid word
    TRIG_BIT_DATA_MICHEL            =0x00000040,   // michel electron, mid word
    TRIG_BIT_DATA_SNXTRIG           =0x00000080,   // The other detector has seen a SN
    TRIG_BIT_DATA_MICHEL_VAR2       =0x00000100,   // Michel electron trigger (production)
    TRIG_BIT_DATA_MICHEL_VAR3       =0x00000200,   // Michel electron trigger variant 2
    TRIG_BIT_DATA_MICHEL_VAR4       =0x00000400,   // Michel electron trigger variant 3
    TRIG_BIT_DATA_MICHEL_TEST       =0x00000800,   // Michel electron test trigger (test)
    TRIG_BIT_DATA_FASTMONO_NN       =0x00001000,   // Fast Monopole using Neural Network (production)
    TRIG_BIT_DATA_FASTMONO_NN_VAR2  =0x00002000,   // Fast Monopole using Neural Network Variant 2 (production)
    TRIG_BIT_DATA_FASTMONO_NN_VAR3  =0x00004000,   // Fast Monopole using Neural Network Variant 3 (production)
    TRIG_BIT_DATA_FASTMONO_NN_TEST  =0x00008000,   // Fast Monopole test trigger using Neural Network (test)
    TRIG_BIT_MID17                  =0x00010000,   // New Triggers
    TRIG_BIT_MID18                  =0x00020000,   // New Triggers
    TRIG_BIT_MID19                  =0x00040000,   // New Triggers
    TRIG_BIT_MID20                  =0x00080000,   // New Triggers
    TRIG_BIT_MID21            =0x00100000,   // New Triggers
    TRIG_BIT_MID22            =0x00200000,   // New Triggers
    TRIG_BIT_MID23            =0x00400000,   // New Triggers
    TRIG_BIT_MID24            =0x00800000,   // New Triggers
    TRIG_BIT_MID25            =0x01000000,   // New Triggers
    TRIG_BIT_MID26            =0x02000000,   // New Triggers
    TRIG_BIT_MID27            =0x04000000,   // New Triggers
    TRIG_BIT_MID28            =0x08000000,   // New Triggers
    TRIG_BIT_MID29            =0x10000000,   // New Triggers
    TRIG_BIT_MID30            =0x20000000,   // New Triggers
    TRIG_BIT_MID31            =0x40000000,   // New Triggers
    TRIG_BIT_MID32            =0x80000000    // New Triggers


  };

  // Trigger Definition Strings
  // These are initialized in 
  extern const char* trigIDStrings[];
  extern const char* trigIDShortStrings[];

  const char* getTriggerName(trigID triggerID);
  const char* getTriggerNameFromMask( unsigned int trigmask);

  const char* getTriggerShortName(trigID triggerID);
  const char* getTriggerShortNameFromMask( unsigned int trigmask);

  trigID    DecodeTrigMask(unsigned int  trigmask );
  trigBitID EncodeTrigMask(trigID   triggerID);

  trigID    DecodeTrigMask32(unsigned int  trigmask );
  trigBitID EncodeTrigMask32(trigID   triggerID);

  trigSourceSubID PartitionToSubID(int32_t);
  int32_t SubIDtoPartition(trigSourceSubID);

} /* endnamespace */

#endif /* TRIGGERDEFINES_H */
