#ifndef DDSMAILBOX__H
#define DDSMAILBOX__H

#include <iostream>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

using namespace gov::fnal::cd::rms;

/**
* @brief DDS mailbox configuration
*/
struct DDSConfig{
        std::string conn_name;
        int rms_partition;
        std::string mbox_name;
};

/**
* @brief Virtual dds mailbox class - a wrapper for DDSConnection
*
* @tparam MessageType - a message class to be sent or received
*/
template <typename MessageType> class DDSMailbox{
 public:    
      /**
      * @brief establish connection with given ConnectionName, partition number and MailboxName
      *
      */
      DDSMailbox(const DDSConfig& cfg):
            _rms_connection (boost::make_shared<provider::DDSConnection>(cfg.conn_name,cfg.rms_partition)),
            _rms_destination(cfg.mbox_name,base::RmsDestination::TRIGGER_CHANNEL) {}

      DDSMailbox(const DDSMailbox<MessageType> &other):
            _rms_connection (other._rms_connection),
            _rms_destination(other._rms_destination)
            {};
      virtual ~DDSMailbox() {};
      ///getters
      const provider::DDSConnection& Connection(){return *_rms_connection;}
      const base::RmsDestination& Destination(){return _rms_destination;}
      
      ///output
      std::string toString(){
        std::string s="DDS Connection:";
        s+="\n- ApplicationName="+Connection().getApplicationName();
        s+="\n- DDS partition="+boost::lexical_cast<std::string>(Connection().getPartitionNumber());
        s+="\nDDS Destination:\n"+Destination().toString();
        return s;
      }
  protected:
    boost::shared_ptr<provider::DDSConnection> _rms_connection; /*< handle for DDSConnection*/
    base::RmsDestination _rms_destination; /*< Connection destination*/
};

/**
* @brief DDSInbox: used for reading messages from given mailbox with >> operator
*
* @tparam MessageType - a message class to be received
*/
template <typename MessageType> class DDSInbox: public DDSMailbox<MessageType>{
  public:
      DDSInbox(const DDSConfig& cfg, unsigned int timeout=100000):
                DDSMailbox<MessageType>(cfg),
                _timeout(timeout),
                _rms_receiver(this->_rms_connection,this->_rms_destination) {};
      DDSInbox(const DDSInbox<MessageType>& other):
                DDSMailbox<MessageType>(other),
                _rms_receiver(this->_rms_connection,this->_rms_destination) {};
      ~DDSInbox() {}
      /**
      * @brief interrupt listening - to stop blocking operation
      */
      void interrupt(){_listening=false;}
      /**
      * @brief receive message
      *
      */
            
      bool receiveMessage(MessageType& message){
            _listening=true;
            while(_listening)
                ///try to receive the message with given 
                if(_rms_receiver.receiveMessage(message,_timeout))
                    return true;
            return false;
      }
           
      DDSInbox<MessageType> & operator >> (MessageType& message){
            receiveMessage(message);
            return *this;
      }

   private:
       unsigned int _timeout;
       RmsReceiver < provider::DDSConnection, MessageType > _rms_receiver;
       bool _listening;
};

/**
* @brief DDSOutbox - sends messages to given Mailbox destination with << operator
*
* @tparam MessageType - a message class to be received
*/
template <typename MessageType> class DDSOutbox: public DDSMailbox<MessageType>{
   public:
      DDSOutbox(const DDSConfig& cfg):
                DDSMailbox<MessageType>(cfg),
                _rms_sender(this->_rms_connection,this->_rms_destination) {};
      DDSOutbox(const DDSOutbox<MessageType> & other):
                DDSMailbox<MessageType>(other),
                _rms_sender(this->_rms_connection,this->_rms_destination) {};

      ~DDSOutbox() {}

      void sendMessage(MessageType& message){
            _rms_sender.sendMessage(message);
      }

      DDSOutbox <MessageType>& operator << (MessageType& message){
            sendMessage(message);
            return *this;
      }
   private:
       RmsSender < provider::DDSConnection, MessageType > _rms_sender;
};

#endif

