#ifndef _RMS_PRODUCER_H
#define _RMS_PRODUCER_H

#include <rms/ClientListenerLoop.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsStatus.h>
#include <rms/provider/BufferedProviderListener.h>
#include <rms/provider/RmsConnection.h>
#include <rms/util/LinkedBlockingQueue.h>

#include <boost/shared_ptr.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#include <vector>
#include <map>
#include <iostream>
#include <set>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

/**
 * This class contains functionality for receiving messages and sending
 * replies.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2010/04/16 18:40:37 $
 */

class RmsProducer : public MessageFilter {
 public:
    RmsProducer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                base::RmsDestination dest);
    RmsProducer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                base::RmsDestination dest, int replyTimeout);
    RmsProducer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                std::vector <base::RmsDestination> dest);
    RmsProducer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                std::vector <base::RmsDestination> dest, int replyTimeout);

    ~RmsProducer();
    
    void sendString(std::string messageText);
    void sendMessage(base::RmsMessage message);
    
    boost::shared_ptr<base::RmsMessage> receiveReply();
    boost::shared_ptr<base::RmsMessage> receiveReply(long timeout);
    boost::shared_ptr<base::RmsMessage> receiveReplyNoWait();
    
    void setReplyListener(RmsMessageListener *listener);
    
    void close();
    
    bool verify(boost::shared_ptr<base::RmsMessage> message);

 private:
    /**
     * Handle to the RMS provider 
     */
    boost::shared_ptr<provider::RmsConnection> _providerConnection; 

    /** 
     * The amount of time to wait for a reply. 
     */
    int _replyTimeout; 
    
    /**
     * Vector of destinations that this producer is sending messages to. 
     */
    std::vector <base::RmsDestination> _myDestinations;

    /**
     * A set of all the channels that this producer is sending messages
     * to, used to reconstruct all the reply destinations.
     */
    std::set <std::string> _replyChannels;

    /**
     * The client listener loop that pulls data off of the input queue and
     * calls our listener 
     */
    ClientListenerLoop _listenerLoop;
    
    /**
     * The input queue that messages are put onto by the provider,
     * pulled off of by the client listener loop. 
     */
    util::LinkedBlockingQueue <boost::shared_ptr<std::string> > _inputQueue; 
    
    /**
     * Handle to the listenerThread that is created in the constructor.
     */
    boost::thread _listenerThread;

    /**
     * Listener that will place new messages onto the input queue.
     */
    provider::BufferedProviderListener *_providerListener;

    /**
     * A table of message UUIDs and dates that is used to keep track 
     * of messages that are sent using this producer to allow the 
     * filtering of replies. 
     */
    std::map <std::string, boost::posix_time::ptime> _replyIdTable; 
    
    /**
     * Perform any initialization that isn't completed in the
     * constructor.
     */
    void initialize();    
};

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
