#ifndef _RMS_EPICSCONNECTION_H
#define _RMS_EPICSCONNECTION_H

#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/provider/RmsConnection.h>
#include <rms/provider/EpicsMessenger.h>
#include <rms/util/UUIDGenerator.h>
#include <rms/util/ReentrantGetEnv.h>

#include <boost/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>

#include <cadef.h>

#include <string>
#include <set>
#include <map>
#include <fstream>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * Manages the interface to EPICS.  Responsibilities include
 * registering listeners with EPICS, creating EpicsMessengers
 * objects when needed to talk to PVs and translating destination
 * objects into PV names.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
class EpicsConnection : public RmsConnection {
 public:
    /**
     * The character that is to be used as a delimiter
     * in PV names.
     */
    static const std::string PV_DELIMITER;

    EpicsConnection (const std::string applicationName,
                     const int applicationPartition);
    ~EpicsConnection ();

    bool supportsDestination(const base::RmsDestination& candidateDestination);
    void sendString(const base::RmsDestination& dest, const std::string& messageString);
    void sendMessage(const base::RmsDestination& dest, const base::RmsMessage& message);
    void addListener(const base::RmsDestination& dest, ProviderListener *listener);
    void removeListener(const base::RmsDestination& dest, ProviderListener *listener);

    bool ping(const base::RmsDestination& pingDestination);

    void close();

 private:
    /**
     * A set of handles to PVs that have been used.
     */
    std::map <std::string, EpicsMessenger *> _cachedPVs;

    /**
     * The EPICS context.
     */
    struct ca_client_context *_caCurrentContext;

    /**
     * Lock for the EPICS code as EpicsMessenger is not thread
     * safe.
     */
    boost::mutex _epicsLock;

    std::string getDestinationPVFromDest(const base::RmsDestination& dest) const;
    std::string getMonitoringPVFromDest(const base::RmsDestination& dest) const;
    EpicsMessenger *getPVHandle(std::string pvName);
};

};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
