#ifndef _RMS_CONNECTION_H
#define _RMS_CONNECTION_H

/**
 * @package gov.fnal.cd.rms.provider
 *
 * @brief Provider-specific classes and interfaces.
 *
 * This package contains classes and interfaces for interacting with
 * third-party messaging providers.
 */

#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/provider/ProviderListener.h>
#include <rms/util/UUIDGenerator.h>

#include <string>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * This is the interface that every provider must
 * implement in order to work with RMS.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2010/02/18 18:27:07 $
 */

class RmsConnection {
 public:
    /**
     * Virtual destructor to suppress compiler warnings.
     */
    virtual ~RmsConnection() { 
        return; 
    }

    /**
     * Determine if this provider supports a particular destination.
     *
     * @param candidateDestination The destination to be checked to see
     * if this provider can send messages to it.
     *
     * @return True if messages can be sent to the destination, false
     * otherwise.
     */
    virtual bool supportsDestination(const base::RmsDestination& 
                                     candidateDestination) = 0;

    /**
     * Send a string to a particular destination.
     *
     * @param dest The destination to send the message to.
     * @param messageString The message to send to the destination.
     */
    virtual void sendString(const base::RmsDestination& dest, 
                             const std::string& messageString) = 0;

    /**
     * Send an RmsMessage to a particular destination.
     *
     * @param dest The destination to send the message to.
     * @param message The message to be sent to the destination.
     */
    virtual void sendMessage(const base::RmsDestination& dest, 
                             const base::RmsMessage& message) = 0;

    /**
     * Add a listener to this provider.  Whenever the provider receives 
     * messages to this destination, the message will be passed to the 
     * listener.
     *
     * @param dest The destination that messages will be going to.
     * @param listener A reference to the listener that will be called
     * when a message is received by the destination.
     */
    virtual void addListener(const base::RmsDestination& dest, 
                             ProviderListener *listener) = 0;

    /**
     * Remove a listener from this provider.
     *
     * @param dest The destination instance to be removed.
     * @param listener The listener instance to be removed.
     */
    virtual void removeListener(const base::RmsDestination& dest,
                                ProviderListener *listener) = 0;

    /**
     * Ping a destination to see if it is up.
     *
     * @return True if the destination is up, false otherwise.
     */
    virtual bool ping(const base::RmsDestination& pingDestination) = 0;

    /**
     * Get a copy of the message source.
     *
     * @return A destination object that contains the source of messages
     * for this connection.
     */
    base::RmsDestination getMessageSource() {
        return _messageSource;
    }

    /**
     * Get a new human readbale UUID from the UUID generator.
     *
     * @return A new UUID.
     */
    const std::string getUUID() {
        return _uuidGenerator->getUUIDHex();
    }

    /**
     * Close down the provider and free up any resources
     * that have been allocated.
     */
    virtual void close() = 0;

 protected:
    /**
     * All providers will need to generate UUIDs.
     */
    util::UUIDGenerator *_uuidGenerator;

    /**
     * A destination to hold the application name and application partition
     * that is passed into the constructor.
     */    
    base::RmsDestination _messageSource;
};

};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
