#ifndef _RMS_RUNNABLE_H
#define _RMS_RUNNABLE_H

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace util {

/**
 * Interface for objects that can be run in separate threads using the
 * Boost Thread library and the RMS ThreadWrapper utility.
 *
 * Sample usage:
 * <code>
 * RunnableDescendant myRunnable;
 * boost::thread
 *   myThread(boost::bind(gov::fnal::cd::rms::util::ThreadWrapper::execute,
 *                        &myRunnable));
 * </code>
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */

// Implementation note:  Classes that implement the Runnable interface
// can make use of the stop() and isRunning() methods *only* if the derived
// classes are careful to keep the relevant attribute values current and
// test against them where appropriate.
//
// In particular, a thorough implementation of the Runnable interface would
// include the following features:
// - assignment of _running to "true" at the beginning of the run() method
// - assignment of _running to "false" at the end of the run() method
// - periodic testing of the _stopRequested attributed within the run()
//   method to see if it should clean up and return

class Runnable {
 public:
    /**
     * Default constructor.
     */
    Runnable() {
        _running = false;
        _stopRequested = false;
    }

    /**
     * Virtual destructor (to ensure that derived class destructors
     * are called).
     */
    virtual ~Runnable() {}

    /**
     * Contains the operations that are to be executed in a separate thread.
     */
    virtual void run() = 0;

    /**
     * Stops the execution of the Runnable object.
     */
    virtual void stop() {
        _stopRequested = true;
    }

    /**
     * Tests if the Runnable object is running.
     *
     * @return true if the object is running, false otherwise.
     */
    virtual bool isRunning() const {
        return _running;
    }

 protected:

    /**
     * Tracks whether the Runnable object is running or not
     * (this can be before it has started or after it has finished)
     */
    bool _running;
    
    /**
     * Indicates that a request has been made to stop running.
     */
    bool _stopRequested;
    
};

};  // end of namespace util

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
