#include "DAQDataFormats/FEBSimulator.h"
#include "DAQDataFormats/TimeStampCounter.h"	// High Resolution Time Stamp Counter

#include <sys/types.h>		// Standard System Types
#include <stdlib.h>		// Standard Library Header
#include <time.h>		// Standard System Time Header

#define FEB_STATUS_SIM 0x40
#define randReal() (1.0*rand() )/RAND_MAX
using namespace daqdataformats;

//------------------------Constructors------------------------------------------
//Default constructor. Use this causiously. FEBLinkID must be defined when generating a nanoslice.
FEBSimulator::FEBSimulator(version_t versionNanoSlice){
  resetSEQ();
  setHitProbability(DEFAULT_HIT_PROBABILITY);	// Default hit probability
  fLastNanoSlice    = new RawNanoSlice(versionNanoSlice);
  fVersionNanoSlice = versionNanoSlice;
}

//------------------------------------------------------------------------------
// Constructor with assigning the FEBLink ID. Main constructor.
FEBSimulator::FEBSimulator(link_id_t linkID,  version_t versionNanoSlice ){
  resetSEQ();
  setLinkID(linkID);
  setHitProbability(DEFAULT_HIT_PROBABILITY);	// Default hit probability
  fLastNanoSlice    = new RawNanoSlice(versionNanoSlice);
  fVersionNanoSlice = versionNanoSlice;
}

//------------------------------------------------------------------------------
// Constructor with assigning the FEB LinkID and a probability. Use this one if you really going to simulate random hits
FEBSimulator::FEBSimulator(link_id_t linkID, float probability, version_t versionNanoSlice){
  resetSEQ();
  setLinkID(linkID);
  setHitProbability(probability);	// Default hit probability
  fLastNanoSlice    = new RawNanoSlice(versionNanoSlice);
  fVersionNanoSlice = versionNanoSlice;
}

//------------------------Destructor--------------------------------------------
FEBSimulator::~FEBSimulator() { if(fLastNanoSlice) delete fLastNanoSlice;}

//------------------------------------------------------------------------------
// Advance the FEB NanoSlice sequence number
void FEBSimulator::advanceSEQ(){
  /// There is no need in masking out. setParameters will take care of that
  fSEQ++;
}

//------------------------------------------------------------------------------
// Reset the FEB NanoSlice sequence number
void FEBSimulator::resetSEQ(){  fSEQ = 0x0;}

//------------------------Setters-----------------------------------------------
void FEBSimulator::setLinkID(link_id_t linkID){  fFEBLinkID = linkID; }

//------------------------------------------------------------------------------
// Set the prob of generating a single hit
void FEBSimulator::setHitProbability(float probability){ fHitProbability = probability;	}

//------------------NanoSlice Generators----------------------------------------
// Default generator. Randomly generates a NanoSlice on a random pixel with a current computer time for TimeStamp
// Might have an empty Nanoslice in the end.
// The output of this method is recorded in fLastNanoSlice.
void FEBSimulator::generateNanoSlice(){

  /// Fill the header information for the nano slice
  /// Pixel can be from 0 to 32
  const uint32_t pixel_address = (uint32_t)rand() & 0x20;

  generateNanoSlice(pixel_address);
}

//------------------------------------------------------------------------------
void FEBSimulator::generateNanoSlice(uint32_t PixelID){

    /// Check to see if we should generate a hit for this block
    if( randReal() < fHitProbability){
        generateNanoSlice(PixelID, true);
        return;
    }

    /// So, there is no hit

    NanoSlice FullNanoSlice;
    // Fill the header information for the nano slice
    FullNanoSlice.Header.DataPresent         = false;
    FullNanoSlice.TimeStamp                  = 0x0;
    FullNanoSlice.PulseHeight                = 0x0;
    FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
    FullNanoSlice.Header.PixelAddress        = PixelID;  // Pixel not random
    FullNanoSlice.Header.FEBStatus           = FEB_STATUS_SIM;
    FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
    FullNanoSlice.Header.DebugMode           = false;
    FullNanoSlice.Header.LinkStatus          = true;
    FullNanoSlice.Header.Version             = (uint32_t)fVersionNanoSlice;

    /// Now Set them parameters
    fLastNanoSlice->setParameters(FullNanoSlice);
}

//------------------------------------------------------------------------------
// This version forces hit generation of a hit on a given pixel
void FEBSimulator::generateNanoSlice(uint32_t PixelID, bool){


  /// Generate the time stamp off the internal clock
  cycle_t full_timestamp;

  full_timestamp.full_time = tsc();
  const uint32_t timeStamp    = full_timestamp.time[0];
  const uint32_t pulse_height = (uint32_t)rand();
  
  generateNanoSlice(PixelID, timeStamp, pulse_height, true);
}

//------------------------------------------------------------------------------
// This version forces hit generation of a hit on a given pixel
void FEBSimulator::generateNanoSlice(uint32_t PixelID, uint32_t iTimeStamp, uint32_t iPulseHeight, bool){

  NanoSlice FullNanoSlice;

  // Fill the header information for the nano slice
  FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
  FullNanoSlice.Header.PixelAddress        = PixelID;
  FullNanoSlice.Header.FEBStatus           = FEB_STATUS_SIM;
  FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
  FullNanoSlice.Header.DebugMode           = false;
  FullNanoSlice.Header.DataPresent         = true;
  FullNanoSlice.Header.LinkStatus          = true;
  FullNanoSlice.Header.Version             = (uint32_t)fVersionNanoSlice;

  FullNanoSlice.TimeStamp                  = iTimeStamp;
  FullNanoSlice.PulseHeight                = iPulseHeight & 0x00000FFF;

  // Now Advance the Sequence Number
  advanceSEQ();
  // Now Set them parameters
  fLastNanoSlice->setParameters(FullNanoSlice);
}

//------------------------------------------------------------------------------
void FEBSimulator::generateNanoSlice(uint32_t PixelID, uint32_t TimeStamp, const std::vector<uint32_t>& data){

    /// Check size of data vector
    if(data.size() == 0)                          return;
    if(fVersionNanoSlice == 1 && data.size() < 3) return;

    /// Pulseheight is valid for both version0 and version1 NanoSlice
    const uint32_t pulse_height = data[0];
    
    generateNanoSlice(PixelID, TimeStamp, pulse_height, true);

    /// Filling other NanoSlice fields
    if(fVersionNanoSlice == 1){
        const uint32_t Value1 = data[1];
        fLastNanoSlice->setValue1(Value1);
    }// end of filling other nanoslice fields

    /// Filling other NanoSlice fields
    if(fVersionNanoSlice == 2){
        const uint32_t Value1 = data[1];
        const uint32_t Value2 = data[2];

        //fLastNanoSlice->setValue1(Value1);
        //fLastNanoSlice->setValue2(Value2);
        fLastNanoSlice->setValue(0, Value1);
        fLastNanoSlice->setValue(1, Value2);
    }// end of filling other nanoslice fields

}


//------------------------------------------------------------------------------
bool FEBSimulator::nonEmptyNanoSlice(){
  if(fLastNanoSlice->getHeader()->getDataPresent()){return true;}
  return false;
}

//------------------------------------------------------------------------------
RawNanoSlice* FEBSimulator::getLastNanoSlice() {return fLastNanoSlice;}
