#include <DatabaseUtils/DAQConfig/GTRunConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

/**
 * Creates a GT run configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 * @param dcmName The name of the DCM of interest.
 */
GTRunConfig::
GTRunConfig(const std::vector<int64_t>& subsystemConfigIdList)
{
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a GT run configuration object from the specified
 * subsystem configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
GTRunConfig::GTRunConfig(const SubsystemNameSpec& nameSpec)
{
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a GT run configuration object from the specified
 * global configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
GTRunConfig::GTRunConfig(const GlobalNameSpec& nameSpec)
{
    _hasValidData=false;
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    if(!NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec,
                                           SubsystemNameSpec::
                                           GT_SUBSYSTEM_NAME,
                                           SubsystemNameSpec::
                                           RUN_CONFIG_TYPE,
                                           subsystemConfigId))
         throw db_exception(SubsystemNameSpec::GT_SUBSYSTEM_NAME,"Could not get ConfigId");

    std::vector<int64_t> subsystemConfigIdList;
    IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
    if(!NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsystemConfigIdList))
        throw db_exception("getConfigIdAncestry"," Failed");

    _primarySubsystemConfigId = subsystemConfigIdList[0];
    _hasValidData = _fetchAllData(subsystemConfigIdList);
}

/**
 * Creates a GT run configuration object from the specified
 * global configuration ID.
 *
 * @param globalConfigId The ID of the global configuration of interest.
 */
GTRunConfig::GTRunConfig(const int64_t& globalConfigId)
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (!NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 GT_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 RUN_CONFIG_TYPE,
                                                 subsystemConfigIdList))
         throw db_exception(SubsystemNameSpec::GT_SUBSYSTEM_NAME,"Could not get ConfigId");
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
}

bool GTRunConfig::
_fetchAllData(const std::vector<int64_t>& subsystemConfigIdList)
{
    std::string fname;

    // for the GT configuration, inheritance is not used,
    // so only the first subsystem ID is relevant
    int64_t subsystemConfigId = subsystemConfigIdList[0];
    std::cout<<"SystemConfigId="<<_primarySubsystemConfigId<<", SubSystemConfigId="<<subsystemConfigId<<std::endl;
        fname = "NovaDatabase/tables/DAQConfig/CalibrationTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_calibrationTriggerList);
        fname = "NovaDatabase/tables/DAQConfig/ExternalTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_externalTriggerList);
        fname = "NovaDatabase/tables/DAQConfig/ManualTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_manualTriggerList);
        fname = "NovaDatabase/tables/DAQConfig/SpillTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_spillTriggerList);
        fname = "NovaDatabase/tables/DAQConfig/DataDrivenTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_dataDrivenTriggerList); 
        fname = "NovaDatabase/tables/DAQConfig/SNEWSTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_SNEWSTriggerList);
        fname = "NovaDatabase/tables/DAQConfig/SuperNovaFilter.xml";
        _fetchTriggerData(fname, subsystemConfigId,_NSNFilterList);
        fname = "NovaDatabase/tables/DAQConfig/DAQStatusTriggers.xml";
        _fetchTriggerData(fname, subsystemConfigId,_daqStatusTriggerList);
        
	fname = "NovaDatabase/tables/DAQConfig/SuperNovaTrigger.xml";
	_fetchParameterSet(fname, subsystemConfigId,_NSNTriggerParams);
	fname = "NovaDatabase/tables/DAQConfig/TriggerOffsets.xml";
        _fetchParameterSet(fname, subsystemConfigId,_triggerOffset);
        fname = "NovaDatabase/tables/DAQConfig/GlobalThrottle.xml";
        _fetchParameterSet(fname, subsystemConfigId,_globalThrottle);
        fname = "NovaDatabase/tables/DAQConfig/DDTThrottle.xml";
        _fetchParameterSet(fname, subsystemConfigId,_DDTThrottle);
        fname = "NovaDatabase/tables/DAQConfig/GTQueue.xml";
        _fetchParameterSet(fname, subsystemConfigId,_gtQueue);
        fname = "NovaDatabase/tables/DAQConfig/SNEWSPipe.xml";
        _fetchParameterSet(fname, subsystemConfigId,_SNEWSPipe);
        fname = "NovaDatabase/tables/DAQConfig/GTGeneral.xml";
        _fetchParameterSet(fname, subsystemConfigId,_gtGeneral);
     return true;
}

bool GTRunConfig::
_fetchParameterSet(const std::string& tableDefinitionFileName,
                  const int64_t& subsystemConfigId,
                  boost::shared_ptr<SimpleParameterSet> & parSet)
{
        TriggerList tmpTriggerList;
        _fetchTriggerData(tableDefinitionFileName, subsystemConfigId,
                                   tmpTriggerList);
        int listSize=tmpTriggerList.size();
	if(listSize==0) throw GTRunConfig::db_exception(tableDefinitionFileName,
	        "ParameterSet is empty");
	if(listSize>1) {
	     char buf[10];
	     sprintf(buf,"%d",listSize);
	     throw GTRunConfig::db_exception(tableDefinitionFileName,
	        "ParameterSet size = "+std::string(buf)+">1");
	}
        parSet = tmpTriggerList[0];
        return true;
}

bool GTRunConfig::
_fetchTriggerData(const std::string& tableDefinitionFileName,
                  const int64_t& subsystemConfigId,
                  TriggerList& triggerList)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(tableDefinitionFileName));
    std::string detName = 
      novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::string dbUser =
      novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {dbt->SetUser(dbUser);}
    if (! dbt->GetConnection()) 
       throw db_exception(tableDefinitionFileName,
                        "Could not connect to DB(detName="+detName+", user="+dbUser+")");
    if (! dbt->ExistsInDB())
       throw GTRunConfig::db_exception(tableDefinitionFileName,
                                    "Table "+dbt->Name()+" doesn't exist in DB");

    dbt->Clear();
    dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                          subsystemConfigId);
    if (! dbt->LoadFromDB())
       throw GTRunConfig::db_exception(tableDefinitionFileName,
                                    "Could not load table "+dbt->Name()+" from DB");

    std::vector<std::string> colNames = dbt->GetColNames();
    NOVADB::Row* rowPtr;
    NOVADB::Column* colPtr;
    std::cout<<"Table "<<dbt->Name()<<"( sscfg="<<subsystemConfigId<<") has "<<dbt->NRow()<<" rows"<<std::endl;
    for (int rowIdx = 0; rowIdx < dbt->NRow(); ++rowIdx) {
        boost::shared_ptr<SimpleParameterSet>
            trigParamSet(new SimpleParameterSet());
        rowPtr = dbt->GetRow(rowIdx);

        for (uint32_t colIdx = 0; colIdx < colNames.size(); ++colIdx) {
            std::string colName = boost::to_lower_copy(colNames[colIdx]);
            if (DBColumnUtils::isInfrastructureColumn(colName)) {continue;}
            if (DBColumnUtils::isSubsystemConfigReferenceColumn(colName)) {
                continue;
            }

            std::string valueString;
            colPtr = rowPtr->Col(colName);
            if (! colPtr->Get(valueString))
              throw GTRunConfig::db_exception(tableDefinitionFileName,
	                                   "Could not read column \""+colName+"\"");
                
           
            trigParamSet->put(colName, valueString);
        }

        triggerList.push_back(trigParamSet);
    }

    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
