#include <iostream>
#include <vector>
#include <NovaDAQUtilities/XMLDeserializationRegistry.h>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/algorithm/string/replace.hpp>

namespace novadaq {

/**
 * Returns an instance of XMLDeserializationRegistry.
 */
XMLDeserializationRegistry&
XMLDeserializationRegistry::getInstance() {
    static XMLDeserializationRegistry singletonInstance;
    return singletonInstance;
}

/**
 * Registers the specified class with the factory.  This method expects
 * a fully qualified class name and an instance of the class (to be used
 * in generating clones as needed).
 *
 * @param className The name of the class that is being registered with
 * the factory.
 * @param classInstance An instance of the class that is being registered.
 */
void XMLDeserializationRegistry::registerClass(std::string className,
                     boost::shared_ptr<XMLSerializable> classInstance) {

    // convert C++-style namespace prefix to Java-style package, if needed.
    // (There is nothing magic about the Java package style, but that is what
    // we chose initially, so we'll use that as the "standard" here.)
    className = translateClassName(className);

    // store the class in the registry
    _registry[className] = classInstance;
}

/**
 * Creates and returns an instance of the requested class.
 *
 * @param className Fully qualified name of the class
 *
 * @return An instance of the class "className"
 */
boost::shared_ptr<XMLSerializable>
XMLDeserializationRegistry::createObject(std::string className) {
    // convert C++-style namespace prefix to Java-style package, if needed.
    className = translateClassName(className);

    // return a clone of the class instance if the class is registered
    if (_registry.find(className) != _registry.end()) {
        boost::shared_ptr<XMLSerializable> classInstance = 
            _registry[className];
        return classInstance->clone();
    }
    else {
        // return an empty pointer if the class is not registered
        boost::shared_ptr<XMLSerializable> emptyPtr;
        return emptyPtr;
    }
}

/**
 * Translates a namespace-qualified class name into a standard format
 * (a::b::c::myClass --> a.b.c.MyClass).
 *
 * @param name The name of the class to translate to the java format.
 */
std::string XMLDeserializationRegistry::translateClassName(std::string name) {
    // replace the double colon with a single period
    boost::replace_all(name, "::", ".");

    // convert the first character in the last token to uppercase
    size_t pos = name.find_last_of(".");
    if (pos != std::string::npos) {
        std::string workString = name;
        std::string targetLetter = workString.substr(pos + 1, 1);
        boost::to_upper(targetLetter);
        name = workString.substr(0, pos + 1) + targetLetter +
            workString.substr(pos + 2);
    }

    // return the modified name
    return name;
}

};  // end of namespace novadaq
