#include <rms/RmsConsumer.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/base/RmsRuntimeException.h>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

/**
 * Create a new consumer given a provider connection and a destination.
 *
 * @param providerConnection A reference to a RMS provider
 * @param dest The destination where we are consuming messages
 * from.
 */
RmsConsumer::RmsConsumer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                         base::RmsDestination dest) :
    _providerConnection(providerConnection),
    _listenerLoop(&_inputQueue),
    _listenerThread(boost::bind(&novadaq::Runnable::run, &_listenerLoop)) {
    dest.validate();
    if (! _providerConnection->supportsDestination(dest)) {
        std::string msg;
        msg.append("The current connection does not support ");
        msg.append("the requested destination: ");
        msg.append(dest.toString());
        GENERATE_RMS_RUNTIME_EXCEPTION(msg);
    }
    _myDestinations.push_back(dest);

    // 02-Nov-2007, KAB - keep track of "verification destinations"
    // explicitly.  These will be used in the verify() method,
    // and they differ slightly depending on whether the destination
    // has the "source" property set.  If they do, then we'll compare
    // to the reply destination in messages.
    std::string targetProperty = 
        _myDestinations[0].getProperty(base::RmsDestination::TARGET_PROPERTY_NAME);
    std::string sourceProperty = 
        _myDestinations[0].getProperty(base::RmsDestination::SOURCE_PROPERTY_NAME);
    if (targetProperty == "" && sourceProperty != "") {
        base::RmsDestination mappedDestination(_myDestinations[0]);
        mappedDestination.setProperty(base::RmsDestination::TARGET_PROPERTY_NAME,
                                      sourceProperty);
        _myVerifyDests.push_back(mappedDestination);
    }
    else {
        _myVerifyDests.push_back(_myDestinations[0]);
    }
  
    _providerListener = new provider::BufferedProviderListener(&_inputQueue);
    _providerConnection->addListener(_myDestinations[0], _providerListener);
    
    _listenerLoop.setMessageFilter(this);
}

/**
 * Create a new consumer given a provider connection and a 
 * vector of destinations.
 *
 * @param providerConnection A reference to a RMS provider
 * @param dest The destination where we are consuming messages
 * from.
 */
RmsConsumer::RmsConsumer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                         std::vector <base::RmsDestination> dest) :
    _providerConnection(providerConnection),
    _listenerLoop(&_inputQueue),
    _listenerThread(boost::bind(&novadaq::Runnable::run, &_listenerLoop)) {

    for (unsigned int i = 0; i < dest.size(); i++) {
        dest[i].validate();
        if (! _providerConnection->supportsDestination(dest[i])) {
            std::string msg;
            msg.append("The current connection does not support ");
            msg.append("the requested destination: ");
            msg.append(dest[i].toString());
            GENERATE_RMS_RUNTIME_EXCEPTION(msg);
        }
    }
    _myDestinations = dest;
  
    _providerListener = new provider::BufferedProviderListener(&_inputQueue);

    for (unsigned int i = 0; i < _myDestinations.size(); i++) {
        _providerConnection->addListener(_myDestinations[i], _providerListener);

        // 02-Nov-2007, KAB - keep track of "verification destinations"
        // explicitly.  These will be used in the verify() method,
        // and they differ slightly depending on whether the destination
        // has the "source" property set.  If they do, then we'll compare
        // to the reply destination in messages.
        std::string targetProperty = 
            _myDestinations[i].getProperty(base::RmsDestination::TARGET_PROPERTY_NAME);
        std::string sourceProperty = 
            _myDestinations[i].getProperty(base::RmsDestination::SOURCE_PROPERTY_NAME);
        if (targetProperty == "" && sourceProperty != "") {
            base::RmsDestination mappedDestination(_myDestinations[i]);
            mappedDestination.setProperty(base::RmsDestination::TARGET_PROPERTY_NAME,
                                          sourceProperty);
            _myVerifyDests.push_back(mappedDestination);
        }
        else {
            _myVerifyDests.push_back(_myDestinations[i]);
        }
    }

    _listenerLoop.setMessageFilter(this);
}

/**
 * Destructor, have close() do all the cleanup.
 */
RmsConsumer::~RmsConsumer() {
    close();
}

/**
 * Receive a string from the provider.  This will block
 * indefinitely.
 *
 * @return A string from the provider
 */
boost::shared_ptr<std::string> RmsConsumer::receiveString() {
    return _inputQueue.take();
}

/**
 * Receive a string from the provider.  This will wait at most
 * timeout seconds.
 *
 * @param timeout The maximum amount of time in seconds to wait
 * before returning.
 *
 * @return A string from the provider if one is available before
 * the timeout expires.  This will return an empty shared_ptr
 * otherwise.
 */
boost::shared_ptr<std::string> RmsConsumer::receiveString(long timeout) {
    return _inputQueue.poll(timeout);
}

/**
 * Receive a string from the provider.  This will not block.
 *
 * @return A string from the provider if one is available.
 * This will return an empty shared_ptr otherwise.
 */
boost::shared_ptr<std::string> RmsConsumer::receiveStringNoWait() {
    return _inputQueue.poll();
}

/**
 * Receive a message from the provider.  This will block
 * indefinitely.
 *
 * @return A message from the provider
 */
boost::shared_ptr<base::RmsMessage> RmsConsumer::receiveMessage() {
    boost::shared_ptr<base::RmsMessage> receivedMessage;
    boost::shared_ptr<std::string> messageText;

    do {
        messageText = _inputQueue.take();
        receivedMessage.reset(new base::RmsMessage);
        receivedMessage->deserialize(*messageText);
    } while (!verify(receivedMessage));
    
    return receivedMessage;
}

/**
 * Receive a message from the provider.  This will wait at most
 * timeout seconds.
 *
 * @param timeout The maximum amount of time in seconds to wait
 * before returning.
 *
 * @return A message from the provider if one is available before
 * the timeout expires.  This will return an empty shared_ptr
 * otherwise.
 */
boost::shared_ptr<base::RmsMessage> RmsConsumer::receiveMessage(long timeout) {
    boost::shared_ptr<base::RmsMessage> receivedMessage;
    boost::shared_ptr<std::string> messageText;

    do {
        messageText = _inputQueue.poll(timeout);
        
        if (!messageText) {
            receivedMessage.reset();
            return receivedMessage;
        }
        
        receivedMessage.reset(new base::RmsMessage);
        receivedMessage->deserialize(*messageText);
    } while (!verify(receivedMessage));

    return receivedMessage;
}

/**
 * Receive a message from the provider.  This will not block.
 *
 * @return A message from the provider if one is available.
 * This will return an empty shared_ptr otherwise.
 */
boost::shared_ptr<base::RmsMessage> RmsConsumer::receiveMessageNoWait() {
    boost::shared_ptr<base::RmsMessage> receivedMessage;
    boost::shared_ptr<std::string> messageText;
    
    do {
        messageText = _inputQueue.poll();
        
        if (!messageText) {
            receivedMessage.reset();
            return receivedMessage;
        }
        
        receivedMessage.reset(new base::RmsMessage);
        receivedMessage->deserialize(*messageText);
    } while (!verify(receivedMessage));
    
    return receivedMessage;
}

/**
 * Set the listener that gets called whenever a message
 * is received.
 *
 * @param listener The new listener
 */
void RmsConsumer::setMessageListener(RmsMessageListener *listener) {
    _listenerLoop.setMessageListener(listener);
}

/**
 * Send a reply to a message.  The message should have been
 * created by calling the createReply() method on the original message.
 *
 * @param message The reply message
 */
void RmsConsumer::sendReply(boost::shared_ptr<base::RmsMessage> message) {
    base::RmsDestination replyDestination;

    replyDestination = _providerConnection->getMessageSource();
    replyDestination.setProperty(base::RmsDestination::CHANNEL_PROPERTY_NAME,
                                 message->getDestination().getProperty(base::RmsDestination::CHANNEL_PROPERTY_NAME));
    replyDestination.validate();
    message->setReplyDestination(replyDestination);

    message->setSentTime();
    message->setId(_providerConnection->getUUID());

    message->getDestination().validate();
    _providerConnection->sendMessage(message->getDestination(), *message);
}

/**
 * Close the consumer and free any resources that it may
 * have allocated.
 */
void RmsConsumer::close() {
    _listenerLoop.stop();
    _listenerThread.join();

    if (_providerListener) {
        delete _providerListener;
        _providerListener = NULL;
    }

    return;
}

/**
 * Verify that a message is relevant to this consumer.  Basically,
 * we just match the destination of the message to the destination
 * we are listening to.  If the message has a status code that isn't
 * success, we send it back to the user even if it doesn't match the
 * consumer's destination.
 *
 * @param message The message to check
 *
 * @return True is the message should be delivered to this consumer,
 * false otherwise.
 */
bool RmsConsumer::verify(boost::shared_ptr<base::RmsMessage> message) {

    // TODO - check if the message status is not success and return "true"
    // if something went wrong so application code learns about the problem.

    std::string messageType = 
        message->getDestination().getProperty(base::RmsDestination::MESSAGE_TYPE_PROPERTY_NAME);

    if (messageType == base::RmsDestination::LOOPBACK_MESSAGE_TYPE) {
        handleLoopback(message);
        return false;
    }

    for (unsigned int i = 0; i < _myVerifyDests.size(); i++) {
        // 02-Nov-2007, KAB - if the consumer destination is a
        // source destination, then compare to the message reply destination
        // rather than the message destination.  (In that case, 
        // we don't care where the message is going *to*, we care where
        // it came *from*).
        std::string sourceProperty =
            _myVerifyDests[i].getProperty(base::RmsDestination::SOURCE_PROPERTY_NAME);
        if (sourceProperty != "") {
            if (message->getReplyDestination().matches(_myVerifyDests[i])) {
                return true;
            }
        }
        else {
            if (message->getDestination().matches(_myVerifyDests[i])) {
                return true;
            }
        }
    }
    
    return false;
}

/**
 * Handle a loopback message that is received in this consumer.
 *
 * @param loopbackMessage The loopback message that has been
 * received.
 */
void RmsConsumer::handleLoopback(boost::shared_ptr<base::RmsMessage> loopbackMessage) {
    base::RmsMessageBody loopbackBody;
    
    loopbackBody.setContent(loopbackMessage->getBody().getContent());
    boost::shared_ptr<base::RmsMessage> loopbackReply = 
        loopbackMessage->createReply();
    loopbackReply->setBody(loopbackBody);
    
    sendReply(loopbackReply);
}

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
