#include <NovaDAQUtilities/XMLDeserializationRegistry.h>
#include <rms/base/RmsMessage.h>
#include <rms/util/TimeUtils.h>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace base {

/**
 * Create an empty RmsMessage.
 */
RmsMessage::RmsMessage() {
}

/**
 * Create an RmsMessage with a given body.
 *
 * @param body The message body to associate with
 * this message.
 */
RmsMessage::RmsMessage(const RmsMessageBody& body) {
    this->body(body);
}

/**
 * Create an RMS message from a serialized string.
 * (This constructor deserializes the input string to
 * populate the attributes of the created instance.)
 *
 * @param serializedData The serialized string that is
 * used to populate this message.
 */
RmsMessage::RmsMessage(std::string serializedData) {
    this->deserialize(serializedData);
}

/**
 * Creates an RMS message with the specified Tranmittable object
 * as the message body.
 *
 * @param bodyObject The XMLSerializable object that will be used
 * as the message body.
 */
RmsMessage::RmsMessage(const novadaq::XMLSerializable& bodyObject) {
    this->setBody(bodyObject);
}

/**
 * Get the UUID that is associated with this message.
 *
 * @return The UUID that is associated with this message
 */
const std::string RmsMessage::getId() const {
    return this->id();
}

/**
 * Get the correlation ID that is associated with this
 * message.  The correlation ID is the UUID of the
 * message to which this message is a reply to.
 *
 * @return The correlation ID of this message.
 */
const std::string RmsMessage::getCorrelationId() const {
    return this->correlationId().get();
}

/**
 * Get the destination that is associated with this message.
 *
 * @return The destination that is associated with this message.
 */
RmsDestination RmsMessage::getDestination() const {
    RmsDestination destination(this->destination());

    return destination;
}

/**
 * Get the destination that replies to this message are to
 * be sent to.
 *
 * @return The destination that replies to this message
 * are to be sent to.
 */
RmsDestination RmsMessage::getReplyDestination() const {
    RmsDestination destination(this->replyDestination().get());

    return destination;
}

/**
 * Get the time at which this message was sent.
 *
 * @return The time this message was sent.
 */
boost::posix_time::ptime RmsMessage::getSentTime() const {
    boost::posix_time::ptime sentTime =
        util::TimeUtils::convertStringToPtime(this->sentTime());

    return sentTime;
}

/**
 * Get the correlation time.  The correlation time
 * is the time the message to which this message
 * is a reply to was sent.
 *
 * @return The correlation time
 */
boost::posix_time::ptime RmsMessage::getCorrelationTime() const {
    boost::posix_time::ptime corrTime =
        util::TimeUtils::convertStringToPtime(this->correlationTime().get());

    return corrTime;
}

/**
 * Get the time to live, in milliseconds, for this message.
 *
 * @return The time to live, in milliseconds, for this message.
 */
long RmsMessage::getTimeToLive() const {
    return this->timeToLive();
}

/**
 * Get the priority of this message.
 *
 * @return The priority of this message.
 */
int RmsMessage::getPriority() const {
    return this->priority().get();
}

/**
 * Get the message source.
 *
 * @return The message source.
 */
RmsMessageSource RmsMessage::getSource() const {
    RmsMessageSource source(this->source().get());

    return source;
}

/**
 * Get the message body.
 *
 * @return The message body.
 */
RmsMessageBody RmsMessage::getBody() const {
    RmsMessageBody body(this->body().get());

    return body;
}

/**
 * Get the message body as a XMLSerializable object.
 *
 * @return A shared pointer to a XMLSerializable object created from
 * the message body.  The shared pointer will be empty
 * if the RmsMessage body was empty.
 */
boost::shared_ptr<novadaq::XMLSerializable> RmsMessage::getBodyObject() const {
    boost::shared_ptr<novadaq::XMLSerializable> bodyObjectPtr;

    // fetch body quantities
    RmsMessageBody ourBody = this->getBody();
    std::string className = ourBody.getClassName();

    // create an instance of the class that is in the body
    novadaq::XMLDeserializationRegistry& registry =
        novadaq::XMLDeserializationRegistry::getInstance();
    bodyObjectPtr = registry.createObject(className);

    // deserialize the body contents into the object
    if (bodyObjectPtr.get() != NULL) {
        bodyObjectPtr->deserialize(ourBody.getContent());
    }

    // return the shared pointer
    return bodyObjectPtr;
}

/**
 * Set the ID of this message.
 *
 * @param uuid An identifier that will uniquely identify
 * this message.
 */
void RmsMessage::setId(const std::string uuid) {
    this->id(uuid);
}

/**
 * Set the correlation ID of this message.  The correlation
 * ID is the ID of the message of which this message is a reply
 * to.
 *
 * @param uuid The correlation ID.
 */
void RmsMessage::setCorrelationId(const std::string uuid) {
    this->correlationId(uuid);
}

/**
 * Set the destination of this message.
 *
 * @param destination The new destination.
 */
void RmsMessage::setDestination(const RmsDestination& destination) {
    this->destination(destination);
}

/**
 * Set the destination that replies to this message
 * will be sent to.
 *
 * @param destination The destination that replies to this
 * message will be sent to.
 */
void RmsMessage::setReplyDestination(const RmsDestination& destination) {
    this->replyDestination(destination);
}

/**
 * Set the time the message was sent to now.
 */
void RmsMessage::setSentTime() {
    boost::posix_time::ptime now = util::TimeUtils::getUniversalPtime();
    this->setSentTime(now);
}

/**
 * Set the time the message was sent.
 *
 * @param timestamp The time the messages was sent.
 */
void RmsMessage::setSentTime(boost::posix_time::ptime timestamp) {
    this->sentTime(util::TimeUtils::convertPtimeToString(timestamp));
}

/**
 * Set the time that the message for which this message is
 * a reply to was sent.
 *
 * @param timestamp The correlation time.
 */
void RmsMessage::setCorrelationTime(boost::posix_time::ptime timestamp) {
    this->correlationTime(util::TimeUtils::convertPtimeToString(timestamp));
}

/**
 * Set the time to live of this message, in milliseconds.
 *
 * @param msec The time to live in milliseconds.
 */
void RmsMessage::setTimeToLive(long msec) {
    this->timeToLive(msec);
}

/**
 * Set the priority of this message.
 *
 * @param value The priority of the message.
 */
void RmsMessage::setPriority(int value) {
    this->priority(value);
}

/**
 * Set the source for this message.
 *
 * @param source The source for this message.
 */
void RmsMessage::setSource(const RmsMessageSource& source) {
    this->source(source);
}

/**
 * Set the body of this message.
 *
 * @param body The body for this message.
 *
 * @bug implement
 */
void RmsMessage::setBody(const RmsMessageBody& body) {
    this->body(body);
}

/**
 * Set the body of this message.
 *
 * @param bodyObject The XMLSerializable object that serves as the body
 * of this message.
 */
void RmsMessage::setBody(const novadaq::XMLSerializable& bodyObject) {
    RmsMessageBody body;
    body.setClassName(novadaq::XMLDeserializationRegistry::
                      translateClassName(bodyObject.getClassName()));
    body.setContent(bodyObject.serialize());
    this->body(body);
}

/**
 * Set the status of this message.
 *
 * @param messageStatus The status of this message.
 */
void RmsMessage::setMessageStatus(const RmsStatus& messageStatus) {
    _messageStatus = messageStatus;
}

/**
 * Get the status of this message.
 *
 * @return The status of this message.
 */
RmsStatus RmsMessage::getMessageStatus() const {
    return _messageStatus;
}

/**
 * Create a reply to this message.  This will return
 * an empty RmsMessage with the destination, reply destination,
 * and correlation fields filled out.
 *
 * <b> Note that the reply destination will be wrong here.  We'll
 * have the provider correct that before it sends the reply. </b>
 *
 * @return An RmsMessage that is setup to be a reply to this
 * message.
 */
boost::shared_ptr<RmsMessage> RmsMessage::createReply() const {
    boost::shared_ptr<RmsMessage> replyMsg (new RmsMessage);
    replyMsg->setCorrelationId(getId());

    RmsDestination newDest = getReplyDestination();
    newDest.setProperty(RmsDestination::MESSAGE_TYPE_PROPERTY_NAME,
                        RmsDestination::REPLY_MESSAGE_TYPE);
    replyMsg->setDestination(newDest);

    replyMsg->setTimeToLive(getTimeToLive());
    return replyMsg;
}

};  // end of namespace base

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
