//  This file (CETDDS.cxx) was created by Ron Rechenmacher <ron@fnal.gov> on
//  Jan 21, 2010. "TERMS AND CONDITIONS" governing this file are in the README
//  or COPYING file. If you do not have such a file, one can be obtained by
//  contacting Ron or Fermi Lab in Batavia IL, 60510, phone: 630-840-3000.
//  $RCSfile: CETDDS.cpp,v $
//  rev="$Revision: 1.10 $$Date: 2011/05/12 17:11:14 $";

#include "rms/provider/CETDDS.h"
using namespace DDS;
#include "rms/provider/CheckStatus.h"
#include "boost/random.hpp"
#include "boost/lexical_cast.hpp"

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

std::string create_extended_topic_name_for( std::string const& type_name,
        std::string const& topic_name )
{
    std::string extendedTopicName(topic_name);
    extendedTopicName.append( "_" );
    extendedTopicName.append( type_name );
    boost::replace_all( extendedTopicName, "::", "_" );
    return extendedTopicName;
}
std::string create_content_filtered_topic_name_for( std::string const& type_name,
        std::string const& topic_name )
{
    return create_extended_topic_name_for(type_name,topic_name.substr(topic_name.find("_")+1));
}

dds_topic_filter_t& create_topic_filter_for(dds_topic_filter_t & filter, std::string const& type_name,
        std::string const& topic_name )
{
    filter.first="( header.target = %0 )";
    filter.second.clear();
    std::string topic_name_tmp(topic_name.substr(0,topic_name.rfind("_")));
    
    //restore dashes replaced in DDSConnection::registerSender
    boost::replace_all(topic_name_tmp,"_dash_","-");
    filter.second.push_back(topic_name_tmp);
        
    return filter;
}

std::string unique_topic_name_for(std::string const& extended_topic_name )
{
    static boost::mt19937 rand(static_cast<unsigned int>(std::time(0)));
    std::string unique_extended_topic_name(extended_topic_name);
    unique_extended_topic_name.append(boost::lexical_cast<std::string>(rand()));
    return  unique_extended_topic_name;
}

bool string2bool (std::string const& bool_string)
{
    return !bool_string.empty() && (
               strcasecmp(bool_string.c_str(),"true")==0
               || atoi(bool_string.c_str()) !=0);
}

void dumpReceivedSamples( dds::SampleInfoSeq const& infoSeq, const unsigned int length, const int status )
{
    std::cout << "++++++++++++++++++++++++++++++++++++++++" << std::endl;
    std::cout << "status=" << status << " topic_seq.length()=" << length << std::endl;

    for ( uint32_t idx = 0; idx < length; ++idx ) {
        std::cout << "sample state = " << infoSeq[idx].sample_state << std::endl;
        std::cout << "view state = " << infoSeq[idx].view_state << std::endl;
        std::cout << "instance state = " << infoSeq[idx].instance_state << std::endl;
        std::cout << "valid data = " << (( int ) infoSeq[idx].valid_data ) << std::endl;
        //std::cout << "source timestamp = " << infoSeq[idx].source_timestamp << std::endl;
        std::cout << "disposed generation count = " << infoSeq[idx].disposed_generation_count << std::endl;
        std::cout << "no writers generation count = " << infoSeq[idx].no_writers_generation_count << std::endl;
        std::cout << "sample rank = " << infoSeq[idx].sample_rank << std::endl;
        std::cout << "generation rank = " << infoSeq[idx].generation_rank << std::endl;
        std::cout << "absolute generation rank = " << infoSeq[idx].absolute_generation_rank << std::endl;
    }

    std::cout << "----------------------------------------" << std::endl;
}

void DPSingleton::reconnect_participant()
{
    disconnect_participant();
    connect_participant();
}

void DPSingleton::disconnect_participant()
{
    dp.reset();

#if CETDDS_DEBUG1
    std::cout << "Disconnected DDS domain participant" << std::endl;
#endif
}

dds::DomainParticipant& DPSingleton::get_participant()
{
    try
    {
        ::DDS::Time_t  current_time;
        
        if(dp.get()==NULL)
          GENERATE_RMS_NOTCONNECTED_EXCEPTION("DDS domain participant is not connected.");

        dp->get()->get_current_time(current_time);

        return *dp;
    }
    catch (...)
    {
        disconnect_participant();
        #if CETDDS_DEBUG1
        std::cout << "DDS domain participant is not connected." << std::endl;
        #endif
        GENERATE_RMS_NOTCONNECTED_EXCEPTION("DDS domain participant is not connected.");
    }
}

void DPSingleton::connect_participant()
{
    ReturnCode_t            status;

    dp.reset(new dds::DomainParticipant);

    /* Tailor a qos policy and make it the new default. */
    dds::TopicQos reliable_topic_qos(*dp);

    status = dp->get()->get_default_topic_qos( reliable_topic_qos );
    checkStatus( status, "dds::DomainParticipant::get_default_topic_qos" );

    reliable_topic_qos.set_reliable();
    reliable_topic_qos.set_volatile();

    reliable_topic_qos.set_keep_last(novadaq::EnvVarCache::getInstance().getEnvVarAsLong("RMS_SAMPLE_HISTORY_SIZE",10));

    status = dp->get()->set_default_topic_qos( reliable_topic_qos );

    checkStatus( status, "dds::DomainParticipant::set_default_topic_qos" );

#if CETDDS_DEBUG1
    std::cout << "Connected DDS domain participant" << std::endl;
#endif
}

DPSingleton::DPSingleton():
        dp(),signalHandler(new ProcessSignalHandler)

{
   try
   {
   connect_participant();
   }catch(...)
   {
     disconnect_participant();
     std::cerr << "DDS domain participant is not connected." << std::endl;
   }
   
   if (!signalHandler->setupSignalHandler(SIGINT))
        std::cerr << "Unable to setup signalHandler for SIGINT in DPSingleton class." << std::endl;

   if (!signalHandler->setupSignalHandler(SIGTERM))
        std::cerr << "Unable to setup signalHandler for SIGTERM in DPSingleton class." << std::endl;

   if (!signalHandler->setupSignalHandler(SIGUSR1))
        std::cerr << "Unable to setup signalHandler for SIGUSR1 in DPSingleton class." << std::endl;

   if (!signalHandler->setupSignalHandler(SIGSEGV))
     std::cerr << "Unable to setup signalHandler for SIGSEGV in DPSingleton class." << std::endl;
   
}
void DPSingleton::addToCleanupList(base::Notifiable* notifiable)
{
  assert( notifiable != NULL );

  signalHandler->addToNotificationList (notifiable);
}

void DPSingleton::removeFromCleanupList(base::Notifiable* notifiable)
{
  assert( notifiable != NULL );

  signalHandler->removeFromNotificationList(notifiable);
}


};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
