#include <rms/provider/MessageAssembler.h>

#include <iostream>
#include <cstring>  /* memcpy */

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * Creates a new assembler instance with the specified initial fragment.
 *
 * @param initialFragment The first fragment that has been received for
 * this message.
 */
MessageAssembler::MessageAssembler(MessageFragment *initialFragment) {
    _requiredFragmentCount = initialFragment->getFragmentCount();

    _messageId = new char[16];
    memcpy(_messageId, initialFragment->getMessageId(), 16);

    _fragments.push_back(initialFragment);
}

/**
 * Destructor, make sure to free up all message
 * fragments.
 */
MessageAssembler::~MessageAssembler() {
    for (unsigned int i = 0; i < _fragments.size(); i++) {
        delete _fragments[i];
    }

    if (_messageId) {
        delete[] _messageId;
        _messageId = NULL;
    }
}

/**
 * Add a message fragment to the assembler.
 *
 * @param fragment The fragment to add
 *
 * @return False if the fragment doesn't belong to this message,
 * or if this fragment has already been received, or any other
 * information in this fragment conflicts with the initial fragment.
 * Returns true if a fragment has been successfully added.
 */ 
bool MessageAssembler::addFragment(MessageFragment *fragment) {
    int fragmentIndex;

    if (!fragment) {
        return false;
    }

    /**
     * The Java version of RMS will only attach UUIDs to messages
     * that are composed of multiple fragments.  If we don't have
     * a UUID, we just need to check to make sure that the message
     * only consists of one fragment.
     */
    if (_messageId[0] != '\0') {
        if (memcmp(fragment->getMessageId(), _messageId, 16) != 0) {
            return false;
        }
    }
    else if (fragment->getFragmentCount() != 1) {
        return false;
    }

    if (fragment->getFragmentCount() != _requiredFragmentCount) {
        return false;
    }

    fragmentIndex = fragment->getFragmentIndex();

    if (fragmentIndex < 0 || fragmentIndex >= _requiredFragmentCount) {
        return false;
    }

    for (unsigned int i = 0; i < _fragments.size(); i++) {
        if (_fragments[i]->getFragmentIndex() == fragmentIndex) {
            return false;
        }
    }
    
    _fragments.push_back(fragment);

    return true;
}

/**
 * Determine if this assembler has all the fragments
 * necessary to assemble a complete message.
 *
 * @return True if this assembler has a complete message.
 */
bool MessageAssembler::hasCompleteMessage() const {
    return ((unsigned int)_requiredFragmentCount == _fragments.size());
}

/**
 * Return a string containing the complete and assembled
 * message
 *
 * @return A string containing the message on success, an
 * empty string if there are fragments missing.
 */
std::string MessageAssembler::getMessage() const {
    std::string message;
    int fragmentsAssembled = 0;

    if (!hasCompleteMessage()) {
        return std::string("");
    }

    while (fragmentsAssembled < _requiredFragmentCount) {
        for (int i = 0; i < _requiredFragmentCount; i++) {
            if (_fragments[i]->getFragmentIndex() == fragmentsAssembled) {
                message.append(_fragments[i]->getPayload());
                fragmentsAssembled++;
            }
        }
    }

    return message;
}
  
};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
