#include <rms/provider/MessageFragment.h>
#include <netinet/in.h>
#include <cstring>  /* memcpy */

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * Special prefix string that is used for all fragments to help
 * tag byte streams as coming from this class (or decodable by
 * this class).
 */
const std::string MessageFragment::_FRAGMENT_PREFIX("RMS");

/**
 * Creates a message fragment with the specified UUID, index, and
 * total fragment count.  The initial payload is empty.
 *
 * @param guid The UUID of the message fragment
 * @param index The index of this fragment into the message, zero based
 * @param totalCount The total number of fragments in the message
 */
MessageFragment::MessageFragment(const char *guid, int index, int totalCount) {
    _uuid = new char[16];
    memcpy(_uuid, guid, 16);

    _fragmentIndex = index;
    _fragmentCount = totalCount;
    
    _messageFragmentPayload = std::string("");
}

/**
 * Creates a message fragment from the specified byte array.  This
 * constructor automatically deserializes the bytes into the
 * appropriate fragment fields.
 *
 * @param fragment The message fragment that we are deserializing.
 */
MessageFragment::MessageFragment(char *fragment) {
    struct fragmentHeader *header;
    _uuid = new char[16];

    header = (struct fragmentHeader *)fragment;

    _fragmentIndex = ntohs(header->fragmentIndex);
    _fragmentCount = ntohs(header->fragmentCount);

    if (header->headerVersion == 0) {
        memcpy(_uuid, header->uuid, 16);
        _messageFragmentPayload.insert(0, fragment + 
                                       sizeof(struct fragmentHeader),
                                       ntohl(header->payloadSize)); 
    }
    else {
        _uuid[0] = '\0';
        _messageFragmentPayload.insert(0, fragment + 
                                       sizeof(struct fragmentHeader) - 
                                       16 * sizeof(char), 
                                       ntohl(header->payloadSize));
    }
}

/**
 * MessageFragment destructor.
 */
MessageFragment::~MessageFragment() {
    if (_uuid) {
        delete[] _uuid;
        _uuid = NULL;
    }
}

/**
 * Returns the index of this fragment (zero-based).
 *
 * @return The index of this fragment.
 */
int MessageFragment::getFragmentIndex() const {
    return _fragmentIndex;
}

/**
 * Returns the total number of fragments in the message
 * that this fragment is a part of.
 *
 * @return The fragment count
 */
int MessageFragment::getFragmentCount() const {
    return _fragmentCount;
}

/**
 * Returns the UUID of the message.
 *
 * @return The UUID of the message.
 */
char *MessageFragment::getMessageId() const {
    return _uuid;
}

/**
 * Returns the size of the message payload, in bytes.
 *
 * @return The size of the message payload.
 */
int MessageFragment::getPayloadSize() const {
    return _messageFragmentPayload.length();
}

/**
 * Set the message payload for this fragment.
 *
 * @param payloadString The payload for this fragment.
 */  
void MessageFragment::setPayload(std::string payloadString) {
    _messageFragmentPayload = payloadString;

    return;
}

/**
 * Get the payload contained inside this fragment.
 *
 * @return The fragment payload.
 */
std::string MessageFragment::getPayload() const {
    return _messageFragmentPayload;
}

/**
 * Serialize the fragment, including all the necessary header
 * information into a character array.
 *
 * @return The serialized message fragment.
 */
void MessageFragment::getBytes(char **message, int *size) const {
    char *fragment;
    struct fragmentHeader *header;

    fragment = new char[sizeof(struct fragmentHeader) + 
                        _messageFragmentPayload.length()];
    header = (struct fragmentHeader *) fragment;

    strncpy(header->fragmentPrefix, _FRAGMENT_PREFIX.c_str(), 3);
    header->headerVersion = 0;
    header->fragmentIndex = htons(_fragmentIndex);
    header->fragmentCount = htons(_fragmentCount);
    header->payloadSize = htonl(_messageFragmentPayload.length());
    memcpy(header->uuid, _uuid, 16);

    strncpy(fragment + sizeof(struct fragmentHeader), 
            _messageFragmentPayload.c_str(), 
            _messageFragmentPayload.length());

    *message = fragment;
    *size = sizeof(struct fragmentHeader) + _messageFragmentPayload.length();
}

/**
 * Clear out the message payload.
 */
void MessageFragment::clearPayload() {
    _fragmentCount = 0;
    _fragmentIndex = 0;
    _uuid[0] = '\0';
    _messageFragmentPayload = "";

    return;
}

/**
 * Return the maximum size of a message header.
 *
 * @return The maximum size of a message header.
 */
int MessageFragment::getMaxHeaderSize() {
    return sizeof(struct fragmentHeader);
}
  
};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
