#include <rms/util/UUIDGenerator.h>

#include <iostream>
#include <cstring>  /* memcpy */

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace util {

/**
 * UUID constructor.  Takes the nodename, and generates the
 * hash.  Also sets up the clock sequence.
 *
 * @param nodeName The name of the node generating the UUIDs.
 *                 This should be unique.
 */
UUIDGenerator::UUIDGenerator(std::string nodeName) {
    _nodeName = nodeName;
    sha1_forRMS_context ctx;
    unsigned int randomNumber;

    gettimeofday(&_previousTime, NULL);

    sha1_forRMS_starts(&ctx);
    sha1_forRMS_update(&ctx, (unsigned char *)_nodeName.c_str(),
                       _nodeName.size());
    sha1_forRMS_finish(&ctx, _nodeNameHash);

    boost::minstd_rand generator(_previousTime.tv_usec);
    boost::uniform_int<> uniformDistribution(0, 65536);
    boost::variate_generator<boost::minstd_rand&, boost::uniform_int<> > 
        uniformDistributionGenerator(generator, uniformDistribution);

    randomNumber = uniformDistributionGenerator() & 0xffff;
    
    _clockSequence[0] = randomNumber & 0xff;
    _clockSequence[1] = (randomNumber >> 8) & 0xff;
    
    return;
}

/**
 * Get a human readable UUID from the generator.
 *
 * @return A universal unique identifier.
 */
const std::string UUIDGenerator::getUUIDHex() {
    unsigned char uuid[16];
    unsigned char uuidHex[37];
    struct timeval currentTime;
    unsigned long timeHigh;
    unsigned long timeLow;
    unsigned long version = 0x1;
    int i = 0;
    int j = 0;

    gettimeofday(&currentTime, NULL);
    
    if (_previousTime.tv_sec == currentTime.tv_sec &&
        _previousTime.tv_usec == currentTime.tv_usec) {
        _overrunCounter++;
    }
    else {
        _overrunCounter = 0;
    }
  
    timeHigh = currentTime.tv_sec;
    timeLow = (currentTime.tv_usec << 10) | (_overrunCounter & 0x3ff);

    timeHigh = timeHigh | (version << 28);

    uuid[8] = _clockSequence[0];
    uuid[9] = _clockSequence[1];

    memcpy(uuid, &timeLow, sizeof(unsigned int));
    memcpy(uuid + (4 * sizeof(char)), &timeHigh, sizeof(unsigned int));

    for (int i = 10; i < 16; i++) {
        uuid[i] = _nodeNameHash[i];
    }

    unsigned int upper, lower;

    for (i = 0; i < 16; i++) {
        upper = (uuid[i] >> 4) & 0x0f;
        lower = uuid[i] & 0x0f;
        
        if (j == 8 || j == 13 || j == 18 || j == 23) {
            uuidHex[j] = '-';
            j++;
        }

        if (upper < 10) {
            uuidHex[j] = upper + 48;
        }
        else {
            uuidHex[j] = upper + 87;
        }

        j++;

        if (lower < 10) {
            uuidHex[j] = lower + 48;
        }
        else {
            uuidHex[j] = lower + 87;
        }
    
        j++;
    }

    uuidHex[36] = '\0';
    const std::string result((const char*)uuidHex);

    return result;
}

/**
 * Get a UUID from the generator.
 *
 * @return A universal unique identifier.
 */
const char *UUIDGenerator::getUUID() {
    char *uuid;
    struct timeval currentTime;
    unsigned long timeHigh;
    unsigned long timeLow;
    unsigned long version = 0x1;

    gettimeofday(&currentTime, NULL);
    
    if (_previousTime.tv_sec == currentTime.tv_sec &&
        _previousTime.tv_usec == currentTime.tv_usec) {
        _overrunCounter++;
    }
    else {
        _overrunCounter = 0;
    }
  
    timeHigh = currentTime.tv_sec;
    timeLow = (currentTime.tv_usec << 10) | (_overrunCounter & 0x3ff);
    
    timeHigh = timeHigh | (version << 28);

    uuid = new char[16];

    uuid[8] = _clockSequence[0];
    uuid[9] = _clockSequence[1];

    memcpy(uuid, &timeLow, sizeof(unsigned int));
    memcpy(uuid + (4 * sizeof(char)), &timeHigh, sizeof(unsigned int));

    for (int i = 10; i < 16; i++) {
        uuid[i] = _nodeNameHash[i];
    }

    // 10-Sep-2009, KAB - ensure that the first byte is not NULL.
    // This is needed because we don't always bother with a UUID in single-
    // fragment messages on the Java side, and the way that we tell if
    // a UUID is empty on the C++ side is to check for a NULL first byte.
    // So, we need to avoid the case in which the full UUID is not NULL
    // but appears to be because its first byte is NULL.
    if (uuid[0] == '\0') {uuid[0] = '\1';}

    return uuid;
}

};  // end of namespace util

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
